//    WeIRCd IRC server software - sub file
//    Copyright (C) 2007-2009 Jonas Thiem
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License v3 as published
//    by the Free Software Foundation.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.

/* moduleslist.c - Extended module state manager */
/* --------------------- */
/* Highlevel module state manager that manages and adjusts module states.
   See moduleslist.h for long description*/

#include "weircd.h"

struct moduleslistentry {
	char* name;
	int normal; //load explicitely as normal module aswell
	int servicesapi; //register for services (startup/shutdown)
	int servicesoptioncount;
	void* servicesoptionnames;
	void* servicesoptionvalues;
};
void* moduleslist = NULL; //settings obtained from the config
void* moduleslist_inuse = NULL; //the one in use, not the one currently obtained by the config


void moduleslist_freeitem(void* item) {
	/* Free all the sub data a list item holds (both config and current state list) */
	struct moduleslistentry* me = linkedlist_GetData(item);
	if (me->name != NULL) {
		free(me->name);
	}
	if (me->servicesoptionnames != NULL) {
		linkedlist_Free(&me->servicesoptionnames);
	}
	if (me->servicesoptionvalues != NULL) {
		linkedlist_Free(&me->servicesoptionvalues);
	}
}
void moduleslist_ritem(void* item) {
	/* Free a single item of the config list with all its data */
	moduleslist_freeitem(item);
	linkedlist_RemoveItem(&moduleslist,item);
}
void modlist_Clear() {
	/* Empty the config list. Gets called when the states have
	   been adjusted as much as possible to the config list and
	   we don't need it anymore. */
	while (moduleslist != NULL) {
		moduleslist_ritem(moduleslist);
	}
}
void* moduleslist_item(const char* name) {
	/* Get an item from the config state list by module name */
	void* item = moduleslist;
	while (item != NULL) {
		struct moduleslistentry* me = linkedlist_GetData(item);
		if (me->name != NULL) {
			if (strcasecmp(me->name,name) == 0) {
				return item;
			}
		}
		item = linkedlist_NextItem(item);
	}
	return NULL;
}
void* moduleslistcurrent_item(const char* name) {
	/* Get an item from the current state list by module name */
	void* item = moduleslist_inuse;
	while (item != NULL) {
		struct moduleslistentry* me = linkedlist_GetData(item);
		if (strcasecmp(me->name,name) == 0) {
			return item;
		}
		item = linkedlist_NextItem(item);
	}
	return NULL;
}
void* moduleslist_add(const char* name) {
	/* Add something to the config state list */
	void* olditem = moduleslist_item(name);
	if (olditem != NULL) {return olditem;}
	struct moduleslistentry me;memset(&me,0,sizeof(me));
	me.name = malloc(strlen(name)+1);
	if (me.name == NULL) {return NULL;}
	strcpy(me.name,name);
	void* returnval = linkedlist_Append(&moduleslist,&me,sizeof(me));
	if  (returnval == NULL) {free(me.name);return NULL;}
	return returnval;
}
int modlist_Check(const char* name) {
	/* Check if a module is in the current state list. This is basically
	   used to indicate if a module is loaded and in use or not. */
	void* item = moduleslistcurrent_item(name);
	if (item == NULL) {return -1;}
	return 0;
}
void* modlist_Get(const char* name) {
	/* Get an item in the config state list and add it when necessary */
	void* item = moduleslist_item(name);
	if (item == NULL) {item = moduleslist_add(name);if (item == NULL) {return NULL;}}
	return item;
}
void modlist_AddAsServices(const char* name) {
	/* Add a module to the config/target state as loaded up services module */
	void* item = modlist_Get(name);if (item == NULL) {return;}
	struct moduleslistentry* me = linkedlist_GetData(item);
	me->servicesapi = 1;
}
void modlist_AddServicesOption(const char* name, const char* option, const char* value) {
	/* Add a startup option to a config state list entry of a module */
	void* item = modlist_Get(name);if (item == NULL) {return;}
	struct moduleslistentry* me = linkedlist_GetData(item);
	linkedlist_Append(&me->servicesoptionnames,option,strlen(option)+1);
	linkedlist_Append(&me->servicesoptionvalues,value,strlen(value)+1);
	me->servicesoptioncount++;
}
void modlist_ApplyChangesOnItem(const struct moduleslistentry* old, struct moduleslistentry* tg) {
	/* Do changes that are necessary to reach the state represented by the target "tg" from
	   the old state "old". The target struct is changed to reflect if the changes worked. */
	if (old->servicesapi == 0) {
		if (tg->servicesapi == 1) {
			//printf("Starting up %s's services functionality\n",old->name);
			char msg[300];
			if (tg->servicesoptioncount > 0) {
				char** optionsarray = (char**)linkedlist_ArrayFromList(tg->servicesoptionnames);
				char** valuesarray = (char**)linkedlist_ArrayFromList(tg->servicesoptionvalues);
				if (servicesmodule_CallStartup(old->name,tg->servicesoptioncount, optionsarray, valuesarray) < 0) {
					snprintf(msg,299,"[MODULES] %s: Services startup failed",old->name);msg[299] = 0;
					IOnotice(msg);printf("%s\n",msg);
					tg->servicesapi = 0;
				}else{
					tg->servicesapi = 1;
				}
				free(optionsarray);
				free(valuesarray);
			}else{
				if (servicesmodule_CallStartup(old->name, 0, NULL, NULL) < 0) {
					snprintf(msg,299,"[MODULES] %s: Services startup failed",old->name);msg[299] = 0;
					IOnotice(msg);printf("%s\n",msg);
					tg->servicesapi = 0;
				}else{
					tg->servicesapi = 1;
				}
			}
		}
	}else{
		if (tg->servicesapi == 0) {
			servicesmodule_CallShutdown(old->name);
		}else{
			// FIXME: check if any services options differ
		}
	}
	if (old->servicesapi != 0 || old->normal != 0) {
		if (tg->servicesapi == 0 && tg->normal == 0) {
			//printf("shutting down some apis\n");
			FakeUsers_ShutdownForModule(old->name);
			ctrackerapi_ClearModule(old->name);
		}
	}
}
void modlist_SafeShutdown(const char* name) {
	/* This gets a zero state and uses it as a target state.
	   That will shut down the module by applying changes to
	   reach that zero state. */
	if (module_ShuttingDown() == 1) {return;} /* Important! Don't do a recursive shutdown inside a shutdown!! (module call restrictions handle this) */
	module_ShuttingDown(); /* Important! We are shutting down, mark this so we don't trigger a shutdown inside this shutdown. */
	void* item = moduleslistcurrent_item(name);
	if (item == NULL) {return;}
	struct moduleslistentry shutdownstate;
	memset(&shutdownstate, 0, sizeof(shutdownstate));
	struct moduleslistentry* me = linkedlist_GetData(item);
	modlist_ApplyChangesOnItem(me,&shutdownstate); /* Reaching the zero state. */
	modman_Unload(me->name); /* Unloading the module, it is no longer of any use. */
	moduleslist_freeitem(item); /* Free the current state item, as the module is now gone. */
	linkedlist_RemoveItem(&moduleslist_inuse, item); /* Remove the empty item from the current state list */
	module_ShutdownDone(); /* Shutdown is done. */
}
int modlist_CanUnload(const struct moduleslistentry* data) {
	if (data->servicesapi == 0 && data->normal == 0) {
		return 0;
	}
	return -1;
}

void modlist_ApplyChanges() {
	/* This triggers an update from the state represented by the current
	   state list (moduleslist_inuse) to the config-induced state (moduleslist).
	   
	   The update is done in three steps:
	   - Remove entries of the config induced state that do nothing (that is,
	     represent a module but no request to start it up)
	   - Remove entries of the current state list that aren't in the config
	     list (unload those modules as the config doesn't want them anymore)
	   - Update and add entries to the current state list by cycling through
	     the config list and comparing with the current state list */

	/* First, remove entries of the target state list that won't do
	   anything. This can happen through the config parser when a module
	   is Enabled=false but added anyway as the config block is opened up
	   and the config parser doesn't now about the disabled state yet. */
	int doneremoving = 0;
	while (doneremoving == 0) {
		doneremoving = 1;
		void* item = moduleslist;
		while (item != NULL) {
			struct moduleslistentry* me = linkedlist_GetData(item);
			if (modlist_CanUnload(me) == 0) {
				moduleslist_freeitem(item);
				linkedlist_RemoveItem(&moduleslist,item);
				doneremoving = 0;
				/* The linked list is now potentially unsafe.
				   Jump out of the list and start it from its
				   beginning so we don't run into segfaults. */
				break;
			}
			item = linkedlist_NextItem(item);
		}
	}
	/* Then, unload modules that we find in the current state list
	   that aren't in the config list (so we want to unload them). */
	doneremoving = 0;
	while (doneremoving == 0) {
		doneremoving = 1;
		void* item = moduleslist_inuse;
		while (item != NULL) {
			struct moduleslistentry* me = linkedlist_GetData(item);
			void* newitem = moduleslist_item(me->name);
			if (newitem == NULL) {
				modlist_SafeShutdown(me->name);
				doneremoving = 0;
				/* The linked list is now potentially unsafe.
				   Jump out of the list and start it from its
				   beginning so we don't run into segfaults. */
				break;
			}
			item = linkedlist_NextItem(item);
		}
	}
	/* And now cycle the config state list and compare
	   with the current state list (search respective items)
	   and see if we can change/add things to update the
	   current state to the desired config-induced state. */
	void* item = moduleslist;
	while (item != NULL) {
		/* This item comes from the config list, representing our desired state: */
		struct moduleslistentry* newstate = linkedlist_GetData(item);
		/* This item comes from the current state list (probably empty): */
		void* citem = moduleslistcurrent_item(newstate->name);
		if (citem == NULL) {
			/* We don't have the item in the current state list, so the module is
			   currently not running at all. Let's try to load it first. */
			if (modman_Load(newstate->name) != 0) {
				item = linkedlist_NextItem(item);
				continue;
			}
			/* Loading worked, so we will now try to "apply changes" on a completely
			   zero'ed state struct (apart from allocated & assigned name) to see if
			   we can initialize it properly: */
			struct moduleslistentry me;
			/* The local struct. We will use this to save the new state while we try
			   to apply it and later on transfer this to the allocated entry in the
			   current state list (in case we succeeded in applying the changes). */
			memset(&me,0,sizeof(me));
			me.name = malloc(strlen(newstate->name)+1);
			if (me.name == NULL) {item = linkedlist_NextItem(item);continue;}
			strcpy(me.name,newstate->name);
			
			/* Add a new entry for our module to the current state list.
			   If that works, we will try to apply changes on our local struct "me" and if that
			   worked aswell, copy the new state to that new entry for our module in the
			   current state list.
			   
			   We could add it to the new state list _after_ trying to start up the module,
			   but we don't want to try to start up the module when we're not even able
			   to add that list entry, so we try that before. */
			void* returnval = linkedlist_Append(&moduleslist_inuse,&me,sizeof(me));
			if (returnval == NULL) {
				/* Even adding failed - free allocated name string and continue */
				free(me.name);
				item = linkedlist_NextItem(item);continue;
			}
			/* Ok, now apply the changes which will start it up */
			modlist_ApplyChangesOnItem(&me, newstate); /* Important: This does _NOT_ change the me struct to reflect the new state.
			                                              The me struct is only used to represent a zero state from which to update to a
								      requested state from the config.
								      The newstate struct IS changed to reflect the state we actually reached now. */
			free(me.name); /* We don't need this anymore. The struct has only been used as a starting point for applying changes as described above and is now useless. */
			if (modlist_CanUnload(newstate) == 0) {
				//module failed to startup something and is now useless.
				//therefore, unload it and remove it again from the current state list
				modman_Unload(newstate->name);
				linkedlist_RemoveItem(&moduleslist_inuse,returnval);
				item = linkedlist_NextItem(item);continue;
			}else{
				memcpy(linkedlist_GetData(returnval), newstate, sizeof(struct moduleslistentry));
				struct moduleslistentry* blu = linkedlist_GetData(returnval);
				/* The config list (target state list) will be free'd after its changes
				   have been applied, but we want to keep our strings and options as
				   the current state list now holds them and uses them: */
				newstate->servicesoptionnames = NULL;
				newstate->servicesoptionvalues = NULL;
				newstate->name = NULL;
			}
			struct moduleslistentry* bla = linkedlist_GetData(returnval);
		}else{
			/* The module is already loaded, but we do want to change things. */
			
			/* First, we apply the changes from the wanted state to the current
			   state to have them done. */
			struct moduleslistentry* curr = linkedlist_GetData(citem);
			modlist_ApplyChangesOnItem(curr, newstate);
			/* Apply changes. "curr" remains _unchanged_, but newstate is changed to the state we actually reached. */
			moduleslist_freeitem(citem); /* Now the changes are done and our state is "newstate". Free the old, unchanged state we started from. */
			memcpy(linkedlist_GetData(citem), newstate, sizeof(struct moduleslistentry)); /* Copy the new state into the old state that we now reached. */
			/* And then, as the state is now transferred to the current state list,
			   NULL those items in the config state list as that list will be free'd
			   and also those items now occupied by the current state list if we
			   don't NULL them here. */
			newstate->servicesoptionnames = NULL;
			newstate->servicesoptionvalues = NULL;
			newstate->name = NULL;
		}
		item = linkedlist_NextItem(item);
	}
}
void modules_ShutdownAll() {
	modlist_Clear();
	modlist_ApplyChanges();
}
int modlist_Print(USER uid, const char* format) {
	int c = 0;
	const char* p = udat(uid)->nickname;
	void* item = moduleslist_inuse;
	while (item != NULL) {
		struct moduleslistentry* me = linkedlist_GetData(item);
		c++;
		char types[300] = "";
		if (me->servicesapi == 1) {
			if (strlen(types) > 0) {strcat(types,", ");}
			strcat(types,"SERVICES");
		}
		if (me->normal == 1) {
			if (strlen(types) > 0) {strcat(types,", ");}
			strcat(types,"GENERIC");
		}
		if (strlen(types) <= 0) {strcpy(types,"ZOMBIE");}
		char msg[500];
		snprintf(msg,499,"%s [%s]",me->name,types);msg[499] = 0;
		usnd(uid,format,p,msg);
		item = linkedlist_NextItem(item);
	}
	return c;
}